#!/usr/bin/env python

from argparse import ArgumentParser
import yaml
import warnings
import multiprocessing
import subprocess
import shlex
import os, sys
import numpy as np

from plotter_style import initializetex, style, legstyle, figsize, despineall
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.ticker import AutoMinorLocator
import matplotlib.image as image
import brewer2mpl

class PlotterOptions:
  def __init__(self):
     self.run_name = "radish_example"
     self.output_folder = None

     self.variations = None
     self.nfiles = 1

     self.order = []
     self.data = None
     self.show_error_band = True
     self.show_ratio = False
     self.normalise_to = 'central'
     self.title = None
     self.diff_show = False
     self.cum_show = False
     self.cum_x_label = None
     self.cum_y_label = None
     self.diff_x_label = "x"
     self.diff_y_label = "y"
     self.log_x = False
     self.log_y = False
     self.xmax = None
     self.xmin = None
     self.ymin = None
     self.ymax = None
     self.ymin_ratio = 0.8
     self.ymax_ratio = 1.2
     self.drawstyle = 'default'
     self.step = None
     self.output = "figure"
     return

  def set_plot_options(self, dict):
     if 'order' in dict:
        self.order = dict.get('order')
     if 'data_file' in dict:
        self.data = dict.get('data_file')
     if 'show_error_band' in dict:
        self.show_error_band = dict.get('show_error_band')
     if 'show_ratio' in dict:
        self.show_ratio = dict.get('show_ratio')
     if 'normalise_ratio_to' in dict:
        self.normalise_to = dict.get('normalise_ratio_to')
     if 'title' in dict:
        self.title = dict.get('title')
     if 'differential' in dict:
        if 'show' in dict.get('differential'):
           self.diff_show = dict.get('differential').get('show')
        if 'x_label' in dict.get('differential'):
           self.diff_x_label = dict.get('differential').get('x_label')
        if 'y_label' in dict.get('differential'):
           self.diff_y_label = dict.get('differential').get('y_label')
     if 'cumulative' in dict:
        if 'show' in dict.get('cumulative'):
           self.cum_show = dict.get('cumulative').get('show')
        if 'x_label' in dict.get('cumulative'):
           self.cum_x_label = dict.get('cumulative').get('x_label')
        if 'y_label' in dict.get('cumulative'):
           self.cum_y_label = dict.get('cumulative').get('y_label')
        self.normalise_to = dict.get('normalise_ratio_to')
     if 'options' in dict:
        if 'set_logx' in dict.get('options'):
           self.log_x = dict.get('options').get('set_logx')
        if 'set_logy' in dict.get('options'):
           self.log_y = dict.get('options').get('set_logy')
        if 'xmin' in dict.get('options'):
           self.xmin = dict.get('options').get('xmin')
           if(self.xmin == 'auto'):
              self.xmin = None
        if 'xmax' in dict.get('options'):
           self.xmax = dict.get('options').get('xmax')
           if(self.xmax == 'auto'):
              self.xmax = None
        if 'ymin' in dict.get('options'):
           self.ymin = dict.get('options').get('ymin')
           if(self.ymin == 'auto'):
              self.ymin = None
        if 'ymax' in dict.get('options'):
           self.ymax = dict.get('options').get('ymax')
           if(self.ymax == 'auto'):
              self.ymax = None
        if 'ymin_ratio' in dict.get('options'):
           self.ymin_ratio = dict.get('options').get('ymin_ratio')
           if(self.ymin_ratio == 'auto'):
              self.ymin_ratio = None
        if 'ymax_ratio' in dict.get('options'):
           self.ymax_ratio = dict.get('options').get('ymax_ratio')
           if(self.ymax_ratio == 'auto'):
              self.ymax_ratio = None
        if 'drawstyle' in dict.get('options'):
           self.drawstyle = dict.get('options').get('drawstyle')
     if (self.drawstyle == 'steps-mid'):
        self.step = 'mid'
     if 'output' in dict:
        self.output = dict.get('output')

  def set_run_options(self, dict):
     if 'run_name' in dict:
        self.run_name = dict.get("run_name")
     if 'output_folder' in dict:
        self.output_folder = dict.get("output_folder")

  def set_scale_options(self, dict):
     if 'variations' in dict:
        self.variations = dict.get("variations")
     if (self.variations == 'none'):
        self.show_error_band = False
        self.nfiles = 1
     elif (self.variations == 'correlated'):
        self.nfiles = 5
     elif (self.variations == 'all'):
        self.nfiles = 9
     else:
        raise ValueError('Possible options for scale_settings are: [none, correlated, all] ')

  def getoutput(self):
     if (not self.output_folder is None):
        os.makedirs(str(self.output_folder),exist_ok=True)
        return str(self.output_folder)+"/"+str(self.output)+".pdf"
     else:
        return str(self.output)

  def getinputname(self, accuracy, scale, suffix):
     if (not self.output_folder is None):
        os.makedirs(str(self.output_folder),exist_ok=True)
        return str(self.output_folder)+"/"+str(self.run_name)+"_"+str(accuracy)+"_"+str(scale)+str(suffix)
     else:
        return str(self.run_name)+str(accuracy)+str(scale)+str(suffix)

def submit_job(cmd):
    """ This runs in a separate thread. """
    #subprocess.call(shlex.split(cmd))  # This will block until cmd finishes
    p = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    return (out, err)

def readfile(filename, accuracy):
   res_values = ['NLL', 'NLLp', 'NNLL', 'NNLLp', 'N3LL']
   fo_values = ['LO', 'NLO', 'NNLO']
   exp_values = ['LOexp', 'NLOexp', 'NNLOexp']
   i = 0
   while True:
      # always check if file exists
      if (os.path.isfile(filename)):
         file = filename
         break
      if  (accuracy in fo_values):
         # look only among matched files
         file = filename.replace("_"+accuracy+"_","_"+res_values[i]+'+'+str(accuracy)+"_")
         if (os.path.isfile(file)):
            break
         file = filename.replace("_"+accuracy+"_","_"+res_values[i]+'_'+str(accuracy)+"_")
         if (os.path.isfile(file)):
            break
         file = filename.replace("_"+accuracy+"_","_"+str(accuracy)+'_'+res_values[i]+"_")
         if (os.path.isfile(file)):
            break
         file = filename.replace("_"+accuracy+"_","_"+str(accuracy)+'+'+res_values[i]+"_")
         if (os.path.isfile(file)):
            break
         else:
            i = i+1
            if(i>4):
               raise ValueError("No " + str(accuracy) + " results among files in run folder. Please check your input card.")
      elif (accuracy in exp_values):
         fo_accuracy = accuracy.replace('exp','')
         file = filename.replace("_"+accuracy+"_","_"+res_values[i]+'+'+str(fo_accuracy)+"_")
         if (os.path.isfile(file)):
            break
         file = filename.replace("_"+accuracy+"_","_"+res_values[i]+'_'+str(fo_accuracy)+"_")
         if (os.path.isfile(file)):
            break
         file = filename.replace("_"+accuracy+"_","_"+str(fo_accuracy)+'_'+res_values[i]+"_")
         if (os.path.isfile(file)):
            break
         file = filename.replace("_"+accuracy+"_","_"+str(fo_accuracy)+'+'+res_values[i]+"_")
         if (os.path.isfile(file)):
            break
         else:
            i = i+1
            if(i>4):
               raise ValueError("No " + str(accuracy) + " results among files in run folder. Please check your input card.")
      elif  (accuracy in res_values):
         # check whether pure resummed result exists
         if (os.path.isfile(filename)):
            file = filename
            break
         # else we need to search among matched files
         file = filename.replace("_"+accuracy+"_","_"+str(accuracy)+"+"+fo_values[i]+"_")
         if (os.path.isfile(file)):
            break
         else:
            i = i+1
            if(i>2):
               raise ValueError("No " + str(accuracy) + " results can be found among files in run folder. Please check your input card.")
      else:
         file = filename
         break
   obs, mat,res,fo,exp = np.loadtxt(file, unpack=True)
   if (accuracy in fo_values):
      return obs, fo
   elif (accuracy in res_values):
      return obs, res
   elif (accuracy in exp_values):
      return obs, exp
   else:
      return obs, mat

def main():
    """
    """

    # read arguments from command line
    parser = ArgumentParser(description='input file')
    parser.add_argument('-i', '--input', dest="infilename",
                        help="input file (yaml format)",
                        metavar="INFILE",
                        required=True)
    parser.add_argument('-n', '--num-runs', dest="max_processes",
                        help="number of processes", type=int,
                        metavar="NPROC", default=1)
    args = parser.parse_args()

    with open(args.infilename, 'r') as stream:
       try:
          dictionary = yaml.safe_load(stream)
       except yaml.YAMLError as exc:
          print(exc)
          quit()

    # run settings variables

    options = PlotterOptions()

    if 'run' in dictionary:
       options.set_run_options(dictionary.get('run'))

    if 'plot' in dictionary:
       options.set_plot_options(dictionary.get('plot'))

    if 'scales' in dictionary:
       options.set_scale_options(dictionary.get('scales'))

    # reading all file

    scale_var_string = []
    if( options.variations == 'none' ):
       scale_var_string.append("xmuFcentral_xmuRcentral_xQcentral")
    elif (options.variations == 'correlated'):
       scale_var_string.append("xmuFcentral_xmuRcentral_xQcentral")
       scale_var_string.append("xmuFup_xmuRup_xQcentral")
       scale_var_string.append("xmuFdown_xmuRdown_xQcentral")
       scale_var_string.append("xmuFcentral_xmuRcentral_xQup")
       scale_var_string.append("xmuFcentral_xmuRcentral_xQdown")
    elif (options.variations == 'all'):
       scale_var_string.append("xmuFcentral_xmuRcentral_xQcentral")
       scale_var_string.append("xmuFup_xmuRup_xQcentral")
       scale_var_string.append("xmuFdown_xmuRdown_xQcentral")
       scale_var_string.append("xmuFup_xmuRcentral_xQcentral")
       scale_var_string.append("xmuFcentral_xmuRup_xQcentral")
       scale_var_string.append("xmuFdown_xmuRcentral_xQcentral")
       scale_var_string.append("xmuFcentral_xmuRdown_xQcentral")
       scale_var_string.append("xmuFcentral_xmuRcentral_xQup")
       scale_var_string.append("xmuFcentral_xmuRcentral_xQdown")

    diff = [[] for x in range(len(options.order))]
    cum = [[] for x in range(len(options.order))]
    diff_norm = []
    cum_norm = []
    obsdiff =  []
    obscum =  []
    for i in range(len(scale_var_string)):
       for j in range(len(options.order)):
          if (options.diff_show == True):
             filename = options.getinputname(options.order[j],scale_var_string[i],".diff")
             obs, result = readfile(filename,options.order[j])
             diff[j].append(result)
             obsdiff.append(obs)
             if ( (options.order[j] == options.normalise_to) and (i==0) ):
               diff_norm = diff[j][i]
          if (options.cum_show == True):
             filename = options.getinputname(options.order[j],scale_var_string[i],".cum")
             obs, result = readfile(filename,options.order[j])
             cum[j].append(result)
             obscum.append(obs)
             if ( (options.order[j] == options.normalise_to) and (i==0) ):
               cum_norm = cum[j][i]

    if( (len(diff) > 6 ) or (len(cum) > 6) ):
       raise ValueError('Too many values. The plot will be likely crowded; please choose at most 6 different curves to show.')

    # TODO: ensure only one case is used
    if ( (options.normalise_to == 'data') ):
       if (options.diff_show == True):
          diff_norm.append(options.data)
       if (options.cum_show == True):
          cum_norm.append(options.data)

    diff_up = [[] for x in range(len(options.order))]
    diff_low = [[] for x in range(len(options.order))]
    cum_up = [[] for x in range(len(options.order))]
    cum_low = [[] for x in range(len(options.order))]
    if(options.show_error_band == True):
       for i in range(len(options.order)):
          if (options.diff_show == True):
             diff_up[i]=np.maximum.reduce(diff[i])
             diff_low[i]=np.minimum.reduce(diff[i])
       for i in range(len(options.order)):
          if (options.cum_show == True):
             cum_up[i]=np.maximum.reduce(cum[i])
             cum_low[i]=np.minimum.reduce(cum[i])

    initializetex()
    colorsetdark = brewer2mpl.get_map('Set1', 'qualitative', 6).mpl_colors
    colorsetpastel = brewer2mpl.get_map('Pastel1', 'qualitative', 6).mpl_colors
    facecolor = ['none', 'none', colorsetpastel[2], colorsetpastel[3], colorsetpastel[4]]
    hatch = ('///','\\\\\\',None,None, None, None)
    alpha = (1,1,0.7,0.2,0.2,0.2)
    alphaleg = (None,None,0.7,0.2,0.2)
    dashedlines = [[1,1],[4,2,4,2],[100000,1],[4,1,1,1],[4, 2, 1, 2, 1, 2]]

    #set outputfile and initialize pdfpages
    output = options.getoutput()
    pdf_pages = PdfPages(output)

    if (options.diff_show == True):

       if( options.show_ratio == True):
          gs = gridspec.GridSpec(2, 1, height_ratios=[2.5, 1])
          gs.update(hspace=0.06)
       else:
          gs = gridspec.GridSpec(1, 1,)

       #set size
       fig_dims = []
       fig_dims = figsize(fig_dims)
       fig_dims =[7,6.5]
       fig = plt.figure(figsize=fig_dims)

       ax1 = fig.add_subplot(gs[0])

       if(options.title != None):
          ax1.set_title(r""+options.title)

       if(options.log_x == True):
          ax1.set_xscale('log')
       if(options.log_y == True):
          ax1.set_yscale('log')

       im = image.imread${PLOTTER_SHARE}
       axsmall1 = plt.axes([.80, .80, .1, .1],frameon=False)
       axsmall1.imshow(im, alpha=1, zorder=1,interpolation=None)
       despineall(axsmall1)
       axsmall1.axes.get_xaxis().set_ticks([])
       axsmall1.axes.get_yaxis().set_ticks([])

       style(ax1)
       ax1.yaxis.set_minor_locator(AutoMinorLocator())
       ax1.xaxis.set_minor_locator(AutoMinorLocator())
       ax1.tick_params(which='both', left=True, right=True, bottom=True, direction='in', zorder=100)

       ax1.set_ylabel(options.diff_y_label)
       if (not options.show_ratio):
          ax1.set_xlabel(options.diff_x_label)
       if ( (not options.ymin == None) and (not options.ymax == None) ):
          ax1.set_ylim(bottom=options.ymin,top=options.ymax)

       if ( (not options.xmin == None) and (not options.xmax == None) ):
          ax1.set_xlim(left=options.xmin,right=options.xmax)
       else:
          ax1.set_xlim(obsdiff[0][0],obsdiff[0][-1])

       labels = options.order
       lines = []
       i = 0
       for j in range(len(diff)):
          p2 = None
          p1, = ax1.plot(obsdiff[0], diff[i][0], color=colorsetdark[i],drawstyle=options.drawstyle,lw=2)
          if( options.show_error_band):
             ax1.fill_between(obsdiff[0], diff_low[i], diff_up[i], facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alpha[i],hatch=hatch[i],step=options.step)
          p2, = [plt.Rectangle((0,0),0,0,facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alphaleg[i],hatch=hatch[i])]
          if(not p2 == None ):
              lines += ((p1,p2),)
          else:
             lines += ((p1,),)
          i = i + 1

       leg = ax1.legend(lines,labels,loc='upper left',numpoints=1)
       legstyle(leg)

       if( options.show_ratio == True):
          plt.setp(ax1.get_xticklabels(), visible=False)
          ax2 = fig.add_subplot(gs[1])
          style(ax2)
          ax2.yaxis.set_minor_locator(AutoMinorLocator())
          ax2.xaxis.set_minor_locator(AutoMinorLocator())
          ax2.tick_params(which='both', left=True, right=True, bottom=True, direction='in', zorder=100)
          ax2.set_ylabel(r"Ratio to "+options.normalise_to)
          ax2.set_xlabel(options.diff_x_label)


          if(options.log_x == True):
             ax2.set_xscale('log')

          if ( (not options.ymin_ratio == None) and (not options.ymax_ratio == None) ):
             ax2.set_ylim(bottom=options.ymin_ratio,top=options.ymax_ratio)
          if ( (not options.xmin == None) and (not options.xmax == None) ):
             ax2.set_xlim(left=options.xmin,right=options.xmax)
          else:
             ax2.set_xlim(obsdiff[0][0],obsdiff[0][-1])

          lines = []
          i = 0

          for j in range(len(diff)):
             if (options.normalise_to == 'central'):
                diff_norm = diff[i][0]
             else:
                diff_norm = diff_norm
             p2 = None
             p1, = ax2.plot(obsdiff[0], diff[i][0]/diff_norm, color=colorsetdark[i],drawstyle=options.drawstyle,lw=2)
             if( options.show_error_band):
               ax2.fill_between(obsdiff[0], diff_low[i]/diff_norm, diff_up[i]/diff_norm, facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alpha[i],hatch=hatch[i],step=options.step)
               p2, = [plt.Rectangle((0,0),0,0,facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alphaleg[i],hatch=hatch[i])]
             if(not p2 == None ):
                lines += ((p1,p2),)
             else:
                lines += ((p1,),)
             i = i + 1

	    #save figure
       pdf_pages.savefig(fig,bbox_inches='tight',pad_inches=0.01)

    if (options.cum_show == True):

       if( options.show_ratio == True):
          gs = gridspec.GridSpec(2, 1, height_ratios=[2.5, 1])
          gs.update(hspace=0.06)
       else:
          gs = gridspec.GridSpec(1, 1,)

       #set size
       fig_dims = []
       fig_dims = figsize(fig_dims)
       fig_dims =[7,6.5]
       fig = plt.figure(figsize=fig_dims)

       ax3 = fig.add_subplot(gs[0])

       if(options.title != None):
          ax3.set_title(r""+options.title)

       if(options.log_x == True):
          ax3.set_xscale('log')
       if(options.log_y == True):
          ax3.set_yscale('log')

       im = image.imread${PLOTTER_SHARE}
       axsmall2 = plt.axes([.80, .80, .1, .1],frameon=False)
       axsmall2.imshow(im, alpha=1, zorder=1,interpolation=None)
       despineall(axsmall2)
       axsmall2.axes.get_xaxis().set_ticks([])
       axsmall2.axes.get_yaxis().set_ticks([])

       style(ax3)
       ax3.yaxis.set_minor_locator(AutoMinorLocator())
       ax3.xaxis.set_minor_locator(AutoMinorLocator())
       ax3.tick_params(which='both', left=True, right=True, bottom=True, direction='in', zorder=100)

       ax3.set_ylabel(options.cum_y_label)
       if (not options.show_ratio):
          ax3.set_xlabel(options.cum_x_label)
       if ( (not options.ymin == None) and (not options.ymax == None) ):
             ax3.set_ylim(bottom=options.ymin,top=options.ymax)

       if ( (not options.xmin == None) and (not options.xmax == None) ):
          ax3.set_xlim(left=options.xmin,right=options.xmax)
       else:
          ax3.set_xlim(obscum[0][0],obscum[0][-1])

       labels = options.order
       lines = []
       i = 0
       for j in range(len(cum)):
          p2 = None
          p1, = ax3.plot(obscum[0], cum[i][0], color=colorsetdark[i],drawstyle=options.drawstyle,lw=2)
          if( options.show_error_band):
             ax3.fill_between(obscum[0], cum_low[i], cum_up[i], facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alpha[i],hatch=hatch[i],step=options.step)
          p2, = [plt.Rectangle((0,0),0,0,facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alphaleg[i],hatch=hatch[i])]
          if(not p2 == None ):
              lines += ((p1,p2),)
          else:
             lines += ((p1,),)
          i = i + 1


       leg = ax3.legend(lines,labels,loc='lower right',numpoints=1)
       legstyle(leg)

       if( options.show_ratio == True):
          plt.setp(ax3.get_xticklabels(), visible=False)
          ax4 = fig.add_subplot(gs[1])
          style(ax4)
          ax4.yaxis.set_minor_locator(AutoMinorLocator())
          ax4.xaxis.set_minor_locator(AutoMinorLocator())
          ax4.tick_params(which='both', left=True, right=True, bottom=True, direction='in', zorder=100)
          ax4.set_ylabel("Ratio to "+options.normalise_to)
          ax4.set_xlabel(options.cum_x_label)

          if ( (not options.ymin_ratio == None) and (not options.ymax_ratio == None) ):
             ax4.set_ylim(bottom=options.ymin_ratio,top=options.ymax_ratio)
          if ( (not options.xmin == None) and (not options.xmax == None) ):
             ax4.set_xlim(left=options.xmin,right=options.xmax)
          else:
             ax4.set_xlim(obscum[0][0],obscum[0][-1])

          lines = []
          i = 0

          if(options.log_x == True):
             ax4.set_xscale('log')

          for j in range(len(cum)):
             if (options.normalise_to == 'central'):
                cum_norm = cum[i][0]
             else:
                cum_norm = cum_norm
             p2 = None
             p1, = ax4.plot(obscum[0], cum[i][0]/cum_norm, color=colorsetdark[i],drawstyle=options.drawstyle,lw=2)
             if( options.show_error_band ):
               ax4.fill_between(obscum[0], cum_low[i]/cum_norm, cum_up[i]/cum_norm, facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alpha[i],hatch=hatch[i],step=options.step)
               p2, = [plt.Rectangle((0,0),0,0,facecolor=facecolor[i],lw=1,edgecolor=colorsetdark[i],alpha=alphaleg[i],hatch=hatch[i])]
             if(not p2 == None ):
                lines += ((p1,p2),)
             else:
                lines += ((p1,),)
             i = i + 1

	    #save figure
       pdf_pages.savefig(fig,bbox_inches='tight',pad_inches=0.01)

    #close file
    pdf_pages.close()

    print("Figure saved in "+str(options.getoutput())+" .\n")

if __name__ == '__main__':
    main()
