module geneva_interface
  use types; use consts_dp
  use shower_routines
  use histogram_routines
  use iso_c_binding
  use masters

  implicit none

  public :: book_geneva_histogram, get_geneva_histo, gen_rad_geneva

contains

subroutine book_geneva_histogram(obs, obslength, bin_index_in, scale_index_in, bins_in, nbins) BIND(C)
    use iso_c_binding
    implicit none
    integer(c_int), value                      :: obslength, bin_index_in, scale_index_in, nbins
    character(kind = c_char), intent(in)       :: obs
    real(c_double), dimension(*), intent(in), target :: bins_in
    real(c_double), pointer                          :: bins(:)

    bins => bins_in(1:nbins)
    call book_histogram("geneva", obs(1:obslength), scale_index_in, bins, bin_index_in)
    call allocate_master_integrals("geneva", obs(1:obslength), nbins)

  end subroutine book_geneva_histogram

  subroutine get_geneva_histo(index_in, nbins, obs, res, res2, exp, exp_as, fixed_order, norm) BIND(C)
    use iso_c_binding
    implicit none
    integer(c_int), value, intent(in) :: nbins, fixed_order, index_in
    real(c_double), intent(inout) :: obs(nbins), res(nbins), res2(nbins), exp(nbins), exp_as(nbins), norm
    character(10) :: nbins_string, size_bins
    integer :: i,index
    integer :: n_events

    ! Identify the histogram given the external index
    index = index_in + 1
    ! add 1 as the geneva index starts from 0

    ! Check that the number of bins matches
    if(size(histo_resummed(index)%values%right) .ne. nbins ) then
      write(nbins_string, "(I1)") nbins
      write(size_bins, "(I1)") size(histo_resummed(index)%values%right)
      call wae_error("number of bins does not match: "//nbins_string//"/="//size_bins)
      stop
    endif

    call get_nev(n_events)

    obs(1:nbins) = histo_resummed(index)%values%right(1:nbins)
    res(1:nbins) = histo_resummed(index)%values%weight(1:nbins)/n_events
    res2(1:nbins) = histo_resummed(index)%values%error(1:nbins)/n_events

    exp(1:nbins) = (histo_expanded(index)%values%lo(1:nbins) + histo_expanded(index)%values%nlo(:nbins))
    exp_as(1:nbins) = exp(1:nbins)
    if (fixed_order > 1) then
        exp(1:nbins) = exp(1:nbins) + histo_expanded(index)%values%nnlo(1:nbins)
    end if
    if (fixed_order > 2) then
        exp(1:nbins) = exp(1:nbins) + histo_expanded(index)%values%n3lo(1:nbins)
    end if

    ! normalise the histogram to the lo cross section
    norm = histo_expanded(index)%values%lo(1)

  end subroutine get_geneva_histo

! shower for a single phase-space points
!#######################################################################
  subroutine gen_rad_geneva(pA_in, idA_in, pB_in, idB_in, msqB_in, msqV1_in, msqV2_in, nf_in, jac_in, scale_index_in, bin_index_in, index) BIND(C)
    use iso_c_binding
    integer, value, intent(in) :: nf_in
    real(dp), intent(inout) :: jac_in
    integer, value, intent(in) :: scale_index_in, bin_index_in, index
    integer :: i, j
    real(dp), intent(in), value :: msqB_in, msqV1_in, msqV2_in
    real(dp) :: msqB(-nf_int:nf_int,-nf_int:nf_int),msqV1(-nf_int:nf_int,-nf_int:nf_int)
    real(dp) :: msqV2(-nf_int:nf_int,-nf_int:nf_int)
    real(dp) :: pborn(14,4)
    real(dp) :: pA_in(1:4), pB_in(1:4)
    integer(c_int), value:: idA_in, idB_in
    real(dp) :: shower_wgt
    logical :: debug

    pborn(:,:) = 0._dp
    msqV1(:,:) = 0._dp
    msqV2(:,:) = 0._dp
    msqB(:,:)  = 0._dp
    debug = .false.

    if( nf_in .ne. nf_int ) then
      write(*,*) "Error: mismatch in the number of flavour: nf_in ", nf_in, "different from nf_int ", nf_int
      stop
    endif

    ! assume canonical momenta as input and change to MCFM convention here
    pborn(1,:) = -pA_in(:)
    pborn(2,:) = -pB_in(:)

    if(debug) then
      do i = 1,2
        print *, "particle ",i, "momentum = ",pborn(i,:)
      enddo
    endif

    i = idA_in
    j = idB_in
    if(idA_in == 21) i = 0;
    if(idB_in == 21) j = 0;

    msqB(i,j)  = msqB_in
    msqV1(i,j) = msqV1_in
    msqV2(i,j) = msqV2_in

    ! set the binning flag to true
    call do_binning();
    call set_scale_index(scale_index_in)
    call set_binvar_index(bin_index_in)
    call make_shower(pborn,msqB,msqV1,msqV2,jac_in,shower_wgt,14,index+1)

  end subroutine gen_rad_geneva


end module geneva_interface