!========================================================
!--------------------------------------------------------
! Additional routines for the manipulation of
! PDFs and luminosities, used by the authors
! for internal checks
! -------------------------------------------------------
!========================================================
module pdfs_tools_extra_routines
  use types; use consts_dp
  !! if using LHAPDF, rename a couple of hoppet functions which
  !! would otherwise conflict with LHAPDF
  use hoppet_v1, EvolvePDF_hoppet => EvolvePDF, InitPDF_hoppet => InitPDF
  use ew_parameters
  use rad_tools
  use interpolation_radish
  use pdfs_tools

  implicit none

  public :: lumi_LL_Born, dlumi_NLL_Born, d2lumi_NLL_Born, derivative_Born, inclusive_cross_sections, inclusive_derivative, init_pdfs_hoppet_evolution, init_pdfs_hoppet_evolution_small_alpha
  public :: inclusive_dlumi_NLL, inclusive_init_P_evol, inclusive_dlumi_NLL_loc,  inclusive_dlumi_NNLL_loc, inclusive_lumi_LL, inclusive_lumi_NLL, inclusive_lumi_NNLL
  type(pdf_table),  save, public :: P1_PDFs, P2_PDFs
  type(pdf_table),  save, public :: P0_P1_PDFs, P0_P2_PDFs
  type(pdf_table),  save, public :: P1_P1_PDFs, P1_P2_PDFs, P2_P2_PDFs

  private
  integer  :: nf_pdf

contains
  !=======================================================================================
  !! initialise Hoppet's PDF table by evolution from an LHAPDF input
  !! at some reference scale.
  subroutine init_pdfs_hoppet_evolution(pdf_name, pdf_set, rts, fixed_flavour_scheme)
    character(len=*), intent(in) :: pdf_name
    integer,          intent(in) :: pdf_set
    real(dp),         intent(in) :: rts
    logical,          intent(in) :: fixed_flavour_scheme
    !--------------------------------------------------
    real(dp), pointer :: pdf_Q(:,:)
    real(dp) :: quark_masses(6)
    real(dp) :: Qmax, ref_Q, alphas_Q, alphasPDF
    ! real(dp) :: getqmass, getthreshold ! LHAPDF functions (deprecated)
    interface
       subroutine evolvePDF(x,Q,res)
         use types; implicit none
         real(dp), intent(in)  :: x,Q
         real(dp), intent(out) :: res(*)
       end subroutine evolvePDF
    end interface

    Qmax = max(2e4_dp, rts)
    call init_grid_and_dglap(rts, fixed_flavour_scheme)
    call AllocPDF(grid, pdf_Q)

    ! set up LHAPDF
    call InitPDFsetByName(trim(pdf_name))
    call InitPDF(pdf_set)
    call getNf(nf_pdf)

    ! extract the PDF at our reference (low) Q value
    if (fixed_flavour_scheme) then
       ref_Q  = 2._dp !>> do forward evolution with 5 flavours
    else
       ref_Q  = MZ    !>> do backward evolution across quark thresholds
    end if
    alphas_Q = alphasPDF(ref_Q)
    call InitPDF_LHAPDF(grid, pdf_Q, evolvePDF, ref_Q)

    ! set up a running coupling, set up the PDF tabulation and perform the evolution
    if (fixed_flavour_scheme) then
       !>> fixed flavour number scheme
       call InitRunningCoupling(alfas=alphas_Q, Q=ref_Q, nloop=3, fixnf=nf_pdf)
    else
       !>> variable flavour number scheme
       quark_masses = zero
       !quark_masses(4:6) = (/mc, mb, mt/) !>> TODO: make sure they match the used LHAPDF set [default in ew_parameters]
       quark_masses(4:6) = (/1.51_dp, 4.92_dp, 172.5_dp/) !>> from NNPDF3.1
       call InitRunningCoupling(alfas=alphas_Q, Q=ref_Q, nloop=3, quark_masses=quark_masses)
       call AddNfInfoToPdfTable(PDFs, ash_global)
    end if

    call EvolvePdfTable(PDFs, ref_Q, pdf_Q, dglap_h, ash_global)
    !------------------------------------------------------------
  end subroutine init_pdfs_hoppet_evolution

  !=======================================================================================
  !! initialise Hoppet's PDF table by evolution from an LHAPDF input
  !! with a user-define valued of alphas
  subroutine init_pdfs_hoppet_evolution_small_alpha(pdf_name, pdf_set, ref_Q, alphas_Q, rts)
   character(len=*), intent(in) :: pdf_name
   integer,          intent(in) :: pdf_set
   real(dp),         intent(in) :: ref_Q, alphas_Q, rts
   !--------------------------------------------------
   real(dp), pointer :: pdf_Q(:,:)
   real(dp) :: Qmax
   interface
      subroutine evolvePDF(x,Q,res)
        use types; implicit none
        real(dp), intent(in)  :: x,Q
        real(dp), intent(out) :: res(*)
      end subroutine evolvePDF
   end interface

   Qmax = max(2e4_dp, rts)
   call init_grid_and_dglap(rts)
   call AllocPDF(grid, pdf_Q)

   !------------------------------------------------------------
   ! set up LHAPDF
   call InitPDFsetByName(trim(pdf_name))
   call InitPDF(pdf_set)
   call getNf(nf_pdf)
   ! extract the PDF at our reference Q value
   call InitPDF_LHAPDF(grid, pdf_Q, evolvePDF, ref_Q)

   ! set up a running coupling, set up the PDF tabulation and perform the evolution
   call InitRunningCoupling(alfas=alphas_Q, Q=ref_Q, nloop=3, fixnf=nf_pdf)
   ! Initialise with variable number of active flavours
   !call InitRunningCoupling(alfas=alphas_Q, Q=ref_Q, nloop=3)
   call EvolvePdfTable(PDFs, ref_Q, pdf_Q, dglap_h, ash_global)
   !------------------------------------------------------------
 end subroutine init_pdfs_hoppet_evolution_small_alpha

  !======================================================================
  ! returns the luminosity that enters the cross section at LL, i.e.
  ! evaluated at the default factorization scale for the process;
  function lumi_LL_Born(x1, x2, msqB) result(res)
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp), intent(in)                     :: x1, x2
    real(dp)                                 :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    integer  :: i

    call get_pdfs_Born(cs%muF, x1, x2, cs%collider, pdf1, pdf2)
    res = lumi_Born(pdf1, pdf2, msqB)
  end function lumi_LL_Born

  function dlumi_NLL_Born(muF, x1, x2, msqB) result(res)
    real(dp),                  intent(in) :: muF, x1, x2
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    real(dp) :: P0_conv_pdf1(-6:7), P0_conv_pdf2(-6:7)
    real(dp) :: P1_conv_pdf1(-6:7), P1_conv_pdf2(-6:7)
    real(dp) :: P2_conv_pdf1(-6:7), P2_conv_pdf2(-6:7)
    real(dp) :: P_conv_pdf1(-6:7), P_conv_pdf2(-6:7)
    integer  :: i
    real(dp) :: as2pi

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_conv_pdf1=zero
    P0_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_PDFs, x1, muF, P0_conv_pdf1)
    call EvalPdfTable_xQ(P0_PDFs, x2, muF, P0_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_conv_pdf1 = P0_conv_pdf1/x1
    P0_conv_pdf2 = P0_conv_pdf2/x2
    ! ---
    P1_conv_pdf1=zero
    P1_conv_pdf2=zero
    call EvalPdfTable_xQ(P1_PDFs, x1, muF, P1_conv_pdf1)
    call EvalPdfTable_xQ(P1_PDFs, x2, muF, P1_conv_pdf2)
    ! Hoppet returns x*f(x)
    P1_conv_pdf1 = P1_conv_pdf1/x1
    P1_conv_pdf2 = P1_conv_pdf2/x2
    ! ---
    P2_conv_pdf1=zero
    P2_conv_pdf2=zero
    call EvalPdfTable_xQ(P2_PDFs, x1, muF, P2_conv_pdf1)
    call EvalPdfTable_xQ(P2_PDFs, x2, muF, P2_conv_pdf2)
    ! Hoppet returns x*f(x)
    P2_conv_pdf1 = P2_conv_pdf1/x1
    P2_conv_pdf2 = P2_conv_pdf2/x2

    if (muF>0.5_dp) then
       as2pi = RunningCoupling(muF)/twopi
    else
       as2pi = RunningCoupling(0.5_dp)/twopi
    end if
    call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)

    P_conv_pdf1 = two*as2pi*(P0_conv_pdf1 + as2pi*P1_conv_pdf1 + as2pi**2*P2_conv_pdf1)
    P_conv_pdf2 = two*as2pi*(P0_conv_pdf2 + as2pi*P1_conv_pdf2 + as2pi**2*P2_conv_pdf2)
    res = lumi_Born(pdf1, P_conv_pdf2, msqB) + lumi_Born(P_conv_pdf1, pdf2, msqB)
    return
  end function dlumi_NLL_Born

  function d2lumi_NLL_Born(muF, x1, x2, msqB) result(res)
    real(dp),                  intent(in) :: muF, x1, x2
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    real(dp) :: P0_conv_pdf1(-6:7), P0_conv_pdf2(-6:7)
    real(dp) :: P1_conv_pdf1(-6:7), P1_conv_pdf2(-6:7)
    real(dp) :: P2_conv_pdf1(-6:7), P2_conv_pdf2(-6:7)
    real(dp) :: P_conv_pdf1(-6:7), P_conv_pdf2(-6:7)
    real(dp) :: P0_P0_conv_pdf1(-6:7), P0_P0_conv_pdf2(-6:7)
    real(dp) :: P0_P1_conv_pdf1(-6:7), P0_P1_conv_pdf2(-6:7)
    real(dp) :: P0_P2_conv_pdf1(-6:7), P0_P2_conv_pdf2(-6:7)
    real(dp) :: P1_P1_conv_pdf1(-6:7), P1_P1_conv_pdf2(-6:7)
    real(dp) :: P1_P2_conv_pdf1(-6:7), P1_P2_conv_pdf2(-6:7)
    real(dp) :: P2_P2_conv_pdf1(-6:7), P2_P2_conv_pdf2(-6:7)
    real(dp) :: PP_conv_pdf1(-6:7), PP_conv_pdf2(-6:7)
    integer  :: i
    real(dp) :: as2pi

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_conv_pdf1=zero
    P0_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_PDFs, x1, muF, P0_conv_pdf1)
    call EvalPdfTable_xQ(P0_PDFs, x2, muF, P0_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_conv_pdf1 = P0_conv_pdf1/x1
    P0_conv_pdf2 = P0_conv_pdf2/x2
    ! ---
    P1_conv_pdf1=zero
    P1_conv_pdf2=zero
    call EvalPdfTable_xQ(P1_PDFs, x1, muF, P1_conv_pdf1)
    call EvalPdfTable_xQ(P1_PDFs, x2, muF, P1_conv_pdf2)
    ! Hoppet returns x*f(x)
    P1_conv_pdf1 = P1_conv_pdf1/x1
    P1_conv_pdf2 = P1_conv_pdf2/x2
    ! ---
    P2_conv_pdf1=zero
    P2_conv_pdf2=zero
    call EvalPdfTable_xQ(P2_PDFs, x1, muF, P2_conv_pdf1)
    call EvalPdfTable_xQ(P2_PDFs, x2, muF, P2_conv_pdf2)
    ! Hoppet returns x*f(x)
    P2_conv_pdf1 = P2_conv_pdf1/x1
    P2_conv_pdf2 = P2_conv_pdf2/x2

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_P0_conv_pdf1=zero
    P0_P0_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_P0_PDFs, x1, muF, P0_P0_conv_pdf1)
    call EvalPdfTable_xQ(P0_P0_PDFs, x2, muF, P0_P0_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_P0_conv_pdf1 = P0_P0_conv_pdf1/x1
    P0_P0_conv_pdf2 = P0_P0_conv_pdf2/x2
    ! ---
    P0_P1_conv_pdf1=zero
    P0_P1_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_P1_PDFs, x1, muF, P0_P1_conv_pdf1)
    call EvalPdfTable_xQ(P0_P1_PDFs, x2, muF, P0_P1_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_P1_conv_pdf1 = P0_P1_conv_pdf1/x1
    P0_P1_conv_pdf2 = P0_P1_conv_pdf2/x2
    ! ---
    P0_P2_conv_pdf1=zero
    P0_P2_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_P2_PDFs, x1, muF, P0_P2_conv_pdf1)
    call EvalPdfTable_xQ(P0_P2_PDFs, x2, muF, P0_P2_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_P2_conv_pdf1 = P0_P2_conv_pdf1/x1
    P0_P2_conv_pdf2 = P0_P2_conv_pdf2/x2
    ! ---
    P1_P2_conv_pdf1=zero
    P1_P2_conv_pdf2=zero
    call EvalPdfTable_xQ(P1_P2_PDFs, x1, muF, P1_P2_conv_pdf1)
    call EvalPdfTable_xQ(P1_P2_PDFs, x2, muF, P1_P2_conv_pdf2)
    ! Hoppet returns x*f(x)
    P1_P2_conv_pdf1 = P1_P2_conv_pdf1/x1
    P1_P2_conv_pdf2 = P1_P2_conv_pdf2/x2
    ! ---
    P1_P1_conv_pdf1=zero
    P1_P1_conv_pdf2=zero
    call EvalPdfTable_xQ(P1_P1_PDFs, x1, muF, P1_P1_conv_pdf1)
    call EvalPdfTable_xQ(P1_P1_PDFs, x2, muF, P1_P1_conv_pdf2)
    ! Hoppet returns x*f(x)
    P1_P1_conv_pdf1 = P1_P1_conv_pdf1/x1
    P1_P1_conv_pdf2 = P1_P1_conv_pdf2/x2
    ! ---
    P2_P2_conv_pdf1=zero
    P2_P2_conv_pdf2=zero
    call EvalPdfTable_xQ(P2_P2_PDFs, x1, muF, P2_P2_conv_pdf1)
    call EvalPdfTable_xQ(P2_P2_PDFs, x2, muF, P2_P2_conv_pdf2)
    ! Hoppet returns x*f(x)
    P2_P2_conv_pdf1 = P2_P2_conv_pdf1/x1
    P2_P2_conv_pdf2 = P2_P2_conv_pdf2/x2


    if (muF>0.5_dp) then
       as2pi = RunningCoupling(muF)/twopi
    else
       as2pi = RunningCoupling(0.5_dp)/twopi
    end if
    call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)

    P_conv_pdf1 = two*as2pi*(P0_conv_pdf1 + as2pi*P1_conv_pdf1 + as2pi**2*P2_conv_pdf1)
    P_conv_pdf2 = two*as2pi*(P0_conv_pdf2 + as2pi*P1_conv_pdf2 + as2pi**2*P2_conv_pdf2)
    PP_conv_pdf1 = two*as2pi*(P0_P0_conv_pdf1 + as2pi*P0_P1_conv_pdf1 + as2pi**2*P0_P2_conv_pdf1 + as2pi**2*P1_P1_conv_pdf1 + as2pi**3*P1_P2_conv_pdf1 + as2pi**4*P2_P2_conv_pdf1)
    PP_conv_pdf2 = two*as2pi*(P0_P0_conv_pdf2 + as2pi*P0_P1_conv_pdf2 + as2pi**2*P0_P2_conv_pdf2 + as2pi**2*P1_P1_conv_pdf2 + as2pi**3*P1_P2_conv_pdf2 + as2pi**4*P2_P2_conv_pdf2)

    res = lumi_Born(P_conv_pdf1, P_conv_pdf2, msqB) + &
         & half*(lumi_Born(PP_conv_pdf1, pdf2, msqB) + &
         &       lumi_Born(pdf1, PP_conv_pdf2, msqB))
    return
  end function d2lumi_NLL_Born

  ! returns the numerical order (rank) derivative for the parton luminosity
  ! (used for ptB resummation) It implements a five-point discrete
  ! derivative formula
  function derivative_Born(rank,tt,x1,x2,msqB,msqV1,msqV2,order) result(res)
    implicit none
    real(dp), intent(in) :: tt, x1, x2 !Final scale of the evolution
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int)
    real(dp), intent(in) :: msqV1(-nf_int:nf_int,-nf_int:nf_int), msqV2(-nf_int:nf_int,-nf_int:nf_int)
    integer, intent(in) :: order, rank
    real(dp) :: asb0, histep, res

    asb0 = cs%alphas_muR*beta0
    histep=1d-3 ! initial step for evaluation of numerical derivative

    ! protect for small tt values by adapting the step size
    do
       if (tt-two*histep > zero) exit
       histep = histep/two
    end do

    if (rank == 1) then
       if (order == order_NLL) then
          res = (lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & - 8._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & - lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & + 8._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1))/(12._dp*histep)
       elseif (order == order_NNLL) then
          res = (lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & - 8._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & - lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt+two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & + 8._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt+histep,cs%p),x1,x2,msqB,msqV1,msqV2))/(12._dp*histep)
       elseif (order == order_NNNLL) then
          res = (lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & - 8._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & - lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt+two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & + 8._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt+histep,cs%p),x1,x2,msqB,msqV1,msqV2))/(12._dp*histep)
       end if
    elseif (rank == 2) then
!!! SECOND DERIVATIVE
       if (order == order_NLL) then
          res = (-  1._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & + 16._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & - 30._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & -  1._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1) &
               & + 16._dp*lumi_NLL_Born(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1))/(12._dp*histep**2)
       elseif (order == order_NNLL) then
          res = (-  1._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & + 16._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & - 30._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt,cs%p)),asb0*Ltilde_scalar(tt,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & -  1._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt+two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & + 16._dp*lumi_NNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt+histep,cs%p),x1,x2,msqB,msqV1,msqV2))/(12._dp*histep**2)
       elseif (order == order_NNNLL) then
          res = (-  1._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & + 16._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & - 30._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt,cs%p)),asb0*Ltilde_scalar(tt,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & -  1._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt+two*histep,cs%p),x1,x2,msqB,msqV1,msqV2) &
               & + 16._dp*lumi_NNNLL_Born(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt+histep,cs%p),x1,x2,msqB,msqV1,msqV2))/(12._dp*histep**2)
       end if
    endif


    if (rank > 2) call wae_error("pdfs_tools.f90: Derivative order too high!")

    return

  end function derivative_Born

  !======================================================================
  ! Functions needed for the inclusive case
  !======================================================================

  function inclusive_lumi_LL() result(res)
    real(dp)                                 :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    integer  :: i

    call get_pdfs(cs%muF, cs%collider, pdf1, pdf2)
    res = inclusive_lumi_at_x(pdf1, pdf2)
  end function inclusive_lumi_LL

  function inclusive_lumi_NLL(muF) result(res)
    real(dp),                  intent(in) :: muF(:)
    real(dp)                              :: res(size(muF))
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    integer  :: i

    do i = 1, size(muF)
       call get_pdfs(muF(i), cs%collider, pdf1, pdf2)
       res(i) = inclusive_lumi_at_x(pdf1, pdf2)
    end do
  end function inclusive_lumi_NLL

  function inclusive_dlumi_NLL(muF) result(res)
    real(dp),                  intent(in) :: muF(:)
    real(dp)                              :: res(size(muF))
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    integer  :: i

    do i = 1, size(muF)
       call get_pdfs(muF(i), cs%collider, pdf1, pdf2)
       res(i) = inclusive_lumi_at_x((dglap_h%P_LO .conv. pdf1), pdf2)
       res(i) = res(i)+ inclusive_lumi_at_x(pdf1, (dglap_h%P_LO .conv. pdf2))
    end do
    res = res * two * cs%as2pi
  end function inclusive_dlumi_NLL

  function inclusive_lumi_NNLL(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF(:), lambda(:)
    real(dp)                              :: res(size(muF))
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu
    type(split_mat) :: C_plus_P
    integer  :: i
    type(pdf_rep) :: pdfrep

    call InitSplitMat(C_plus_P, C1_matrix)
    call AddWithCoeff(C_plus_P, dglap_h%P_LO, -cs%ln_muF2_M2+cs%ln_Q2_M2)

    do i = 1, size(muF)
       call get_pdfs(muF(i), cs%collider, pdf1, pdf2)
       ! get the running coupling at scale muR = muR_default * exp(-L)
       as2pi_mu = cs%as2pi / (1-2*lambda(i))

       ! correct the luminosities and get no "spurious" as2pi_mu^2 term
       res(i) = inclusive_lumi_at_x(pdf1, pdf2)
       res(i) = res(i) + as2pi_mu*inclusive_lumi_at_x(C_plus_P .conv. pdf1, pdf2)
       res(i) = res(i) + as2pi_mu*inclusive_lumi_at_x(pdf1, C_plus_P .conv. pdf2)
    end do

    call Delete(C_plus_P)
  end function inclusive_lumi_NNLL

  ! initialise PDF evolution operator (i.e. r.h.s. of DGLAP equation)
  ! The evaluation of this term must fulfil quark-mass thresholds requirements,
  ! which are crossed in the running coupling evolution.
  ! Whenever a threshold is crossed, one has to reset the number of active flavours.
  ! This can be implemented by means of the function SetNfDglapHolder(dglap_h, nf_lcl)
  ! which reset nf in the splitting functions as well as in the QCD objects (e.g. beta function).
  ! The number of flavours in the evolution at a given scale can be accessed via NfAtQ(ash_global, 3._dp, muM_mQ=one).
  ! Here the coupling in the PDF evolution is initialised with a fixed nnumber of quark flavours
  ! (see InitRunningCoupling(alfas=alphasMZ, Q=MZ, nloop=3, fixnf=nf_pdf)) therefore no thresholds are present in
  ! our case
  subroutine inclusive_init_P_evol(muF,P_evol)
    implicit none
    real(dp), intent(in) :: muF !Final scale of the evolution
    type(split_mat), intent(out) :: P_evol
    real(dp) :: as2pi, as_cutoff
    !----------------------------------------------
    ! Use hoppet's running. It has a cutoff set by the tlo variable in
    ! new_as.f90, by default corresponding to a 0.5 GeV cutoff.
    as_cutoff = 0.5d0 ! temporary fix: hoppet's cutoff is 0.5 GeV
    if (muF>as_cutoff) then
       as2pi = RunningCoupling(muF)/twopi
       !write(*,*) NfAtQ(ash_global, 3._dp, muM_mQ=one)
    else
       as2pi = RunningCoupling(as_cutoff)/twopi ! freeze the coupling below the cutoff
    end if
    call InitSplitMat(P_evol, dglap_h%P_LO)
    call AddWithCoeff(P_evol, dglap_h%P_NLO, as2pi)
    call AddWithCoeff(P_evol, dglap_h%P_NNLO, as2pi**2)
    call Multiply(P_evol,two*as2pi)
    return
  end subroutine inclusive_init_P_evol

  !======================================================================
  ! returns the evolution step for the NLL luminosity at NNLO
  function inclusive_dlumi_NLL_loc(muF) result(res)
    real(dp),                  intent(in) :: muF
    type(split_mat) :: P_evol
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    !----------------------------------------------

    call get_pdfs(muF, cs%collider, pdf1, pdf2)
    ! initialise evolution operator
    call inclusive_init_P_evol(muF, P_evol)

    ! use NNLO evolution
    res = inclusive_lumi_at_x((P_evol .conv. pdf1), pdf2)
    res = res + inclusive_lumi_at_x(pdf1, (P_evol .conv. pdf2))

    ! free memory
    call Delete(P_evol)
  end function inclusive_dlumi_NLL_loc

  ! returns the evolution step for the NNLL luminosity at NNLO
  ! (derivative of the NNLL luminosity)
  function inclusive_dlumi_NNLL_loc(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF, lambda
    type(split_mat) :: P_evol
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: C1_x_pdf1(0:grid%ny,-6:7), C1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: C2_x_pdf1(0:grid%ny,-6:7), C2_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: G1_x_pdf1(0:grid%ny,-6:7), G1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu, HH1
    real(dp) :: asb0, hh

    if (cs%loop_mass .ne. 'none') then
       HH1 = Hm(1)
    else
       HH1 = H(1)
    end if

    call get_pdfs(muF, cs%collider, pdf1, pdf2)
    ! initialise evolution operator
    call inclusive_init_P_evol(muF, P_evol)

    ! define useful convolutions
    C1_x_pdf1 = (C1_plus_P .conv. pdf1)
    C1_x_pdf2 = (C1_plus_P .conv. pdf2)
    C2_x_pdf1 = (C2_plus_P .conv. pdf1)
    C2_x_pdf2 = (C2_plus_P .conv. pdf2)
    G1_x_pdf1 = (G1_matrix .conv. pdf1)
    G1_x_pdf2 = (G1_matrix .conv. pdf2)

    ! get the running coupling at scale muR = muR_default * exp(-L)
    as2pi_mu = cs%as2pi / (1-2*lambda)

    res = inclusive_lumi_at_x((P_evol .conv. pdf1), pdf2)
    res = res + inclusive_lumi_at_x(pdf1, (P_evol .conv. pdf2))

    ! add contribution of the virtual corrections piece
    res = res*(one + cs%as2pi*HH1 + cs%as2pi**2*H(2))

    ! add derivatives of the O(as) term:
    ! derivative of the as2pi_mu piece
    ! (minus sign coming from the over all minus sign in front of the luminosity derivative in the master formula)
    res = res - as2pi_mu**2*(four*pi*beta0)*(inclusive_lumi_at_x(C1_x_pdf1, pdf2) &
         & + inclusive_lumi_at_x(pdf1, C1_x_pdf2))*(one + cs%as2pi*HH1)

    ! derivative of (Coefficient function) .conv. luminosity (use commutativity of convolution)
    res = res + as2pi_mu*inclusive_lumi_at_x(P_evol .conv. C1_x_pdf1, pdf2)*(one + cs%as2pi*HH1)
    res = res + as2pi_mu*inclusive_lumi_at_x(C1_x_pdf1, P_evol .conv. pdf2)*(one + cs%as2pi*HH1)
    res = res + as2pi_mu*inclusive_lumi_at_x(pdf1, P_evol .conv. C1_x_pdf2)*(one + cs%as2pi*HH1)
    res = res + as2pi_mu*inclusive_lumi_at_x(P_evol .conv. pdf1, C1_x_pdf2)*(one + cs%as2pi*HH1)


    ! add derivatives of the O(as^2) term:
    ! derivative of the as2pi_mu^2 piece
    ! (minus sign coming from the over all minus sign in front of the luminosity derivative in the master formula)
    res = res - as2pi_mu**3*(8._dp*pi*beta0)*(inclusive_lumi_at_x(C2_x_pdf1, pdf2) &
         & + inclusive_lumi_at_x(pdf1, C2_x_pdf2))

    ! derivative of (Coefficient function) .conv. luminosity (use commutativity of convolution)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(P_evol .conv. C2_x_pdf1, pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(C2_x_pdf1, P_evol .conv. pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(pdf1, P_evol .conv. C2_x_pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(P_evol .conv. pdf1, C2_x_pdf2)

    ! add contribution from the cross products pieces (i.e. C1xC1, G1xG1)
    res = res - as2pi_mu**3*(8._dp*pi*beta0)*(inclusive_lumi_at_x(C1_x_pdf1, C1_x_pdf2) &
         & + inclusive_lumi_at_x(G1_x_pdf1, G1_x_pdf2))

    res = res + as2pi_mu**2*inclusive_lumi_at_x(P_evol .conv. C1_x_pdf1, C1_x_pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(C1_x_pdf1, P_evol .conv. C1_x_pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(P_evol .conv. G1_x_pdf1, G1_x_pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(G1_x_pdf1, P_evol .conv. G1_x_pdf2)
    call Delete(P_evol)
  end function inclusive_dlumi_NNLL_loc

  ! returns the numerical derivative of the parton luminosity (used for ptB resummation)
  ! It implements a five-point discrete derivative formula
  function inclusive_derivative(tt,order) result(res)
    implicit none
    real(dp), intent(in) :: tt !Final scale of the evolution
    integer, intent(in) :: order
    real(dp) :: asb0, histep, res
    asb0 = cs%alphas_muR*beta0
    histep=1d-4 ! step for evaluation of numerical derivative

    ! protect for small tt values by adapting the step size
    do
       if (tt-two*histep > zero) exit
       write(*,*) 'WARNING: tt - 2*histep < 0. Reducing the derivative step'
       histep=histep/10._dp
    end do

    if (order == order_NLL) then
       res = (inclusive_lumi_NLL_loc(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p)) &
            & - 8._dp*inclusive_lumi_NLL_loc(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)), asb0*Ltilde_scalar(tt-two*histep,cs%p))&
            & - inclusive_lumi_NLL_loc(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p)) &
            & + 8._dp*inclusive_lumi_NLL_loc(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p)))/(12._dp*histep)
    elseif (order == order_NNLL) then
       res = (inclusive_lumi_NNLL_loc(cs%muF*exp(-Ltilde_scalar(tt-two*histep,cs%p)),asb0*Ltilde_scalar(tt-two*histep,cs%p)) &
            & - 8._dp*inclusive_lumi_NNLL_loc(cs%muF*exp(-Ltilde_scalar(tt-histep,cs%p)),asb0*Ltilde_scalar(tt-histep,cs%p)) &
            & - inclusive_lumi_NNLL_loc(cs%muF*exp(-Ltilde_scalar(tt+two*histep,cs%p)),asb0*Ltilde_scalar(tt+two*histep,cs%p)) &
            & + 8._dp*inclusive_lumi_NNLL_loc(cs%muF*exp(-Ltilde_scalar(tt+histep,cs%p)),asb0*Ltilde_scalar(tt+histep,cs%p)))/(12._dp*histep)
    end if
    return
  end function inclusive_derivative

  ! ======================================================================
  ! returns an array containing the LO and NLO cross sections
  function inclusive_cross_sections() result(res)
    use coefficient_functions_VH
    real(dp)                              :: res(1:2)
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: P_x_pdf1(0:grid%ny,-6:7), P_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: lumi_gg(0:grid%ny), lumi_qg(0:grid%ny), lumi_qqbar(0:grid%ny)
    integer  :: i
    type(gdval) :: x

    call get_pdfs(cs%muF, cs%collider, pdf1, pdf2)
    x = cs%M2_rts2 .with. grid

    select case(cs%proc)
    case("H")
       call InitHiggsCoeffs(grid)
       lumi_gg = PartonLuminosity(grid, pdf1(:,iflv_g), pdf2(:,iflv_g))
       lumi_qg = PartonLuminosity(grid, pdf1(:,iflv_g), sum(pdf2(:,-6:-1),dim=2)&
            &                                          +sum(pdf2(:, 1:6 ),dim=2) )
       lumi_qg = lumi_qg + &
            &    PartonLuminosity(grid, sum(pdf1(:,-6:-1),dim=2)&
            &                          +sum(pdf1(:, 1:6 ),dim=2), pdf2(:,iflv_g)   )
       lumi_qqbar = 0
       do i = -6, 6
          if (i == 0) cycle
          lumi_qqbar = lumi_qqbar + PartonLuminosity(grid, pdf1(:,i), pdf2(:,-i))
       end do

       res(1) = lumi_gg .atx. x

       res(2) =   ((higgs_H1gg .conv. lumi_gg) .atx. x)&
            &   + ((higgs_H1qg .conv. lumi_qg) .atx. x)&
            &   + ((higgs_H1qqbar .conv. lumi_qqbar) .atx. x)

       if (cs%ln_muF2_M2 /= zero) then
          P_x_pdf1 = dglap_h%P_LO * pdf1
          P_x_pdf2 = dglap_h%P_LO * pdf2
          res(2) = res(2) - (  (PartonLuminosity(grid, P_x_pdf1(:,0), pdf2(:,0)).atx.x)&
               &             + (PartonLuminosity(grid, pdf1(:,0), P_x_pdf2(:,0)).atx.x))*cs%ln_muF2_M2
       end if
       if (cs%ln_muR2_M2 /= zero) then
          res(2) = res(2) + 2*twopi*beta0*cs%ln_muR2_M2 * res(1)
       end if


       res(2) = res(2) * cs%as2pi
       res    = res * cs%norm_fb
    case("DY")
       call InitDYCoeffs(grid)
       lumi_qqbar(0:grid%ny) = inclusive_lumi_all_x(pdf1, pdf2)
       res(1) = lumi_qqbar .atx. x

       res(2) = (DY_D1qg .conv. inclusive_lumi_DY_qg(pdf1, pdf2)) .atx. x
       res(2) = res(2) + ( (DY_D1qqbar * lumi_qqbar) .atx. x)

       if (cs%ln_muF2_M2 /= zero) then
          P_x_pdf1 = dglap_h%P_LO * pdf1
          P_x_pdf2 = dglap_h%P_LO * pdf2
          res(2) = res(2) - (  (inclusive_lumi_all_x(P_x_pdf1, pdf2).atx.x)&
               &             + (inclusive_lumi_all_x(pdf1, P_x_pdf2).atx.x))*cs%ln_muF2_M2
       end if
       ! NB: do nothing for muR scale dependence at this order.

       res(2) = res(2) * cs%as2pi
      case default
        call wae_error("unrecognized process: "//trim(cs%proc))

    end select

  end function inclusive_cross_sections

    !----------------------------------------------------------------------
  function inclusive_lumi_DY_qg(pdf1,pdf2) result(res)
    real(dp),                     intent(in) :: pdf1(:,-6:), pdf2(:,-6:)
    real(dp)                     :: res(size(pdf1,1)), weighted_q
    real(dp)                     :: weighted_q1(size(pdf1,1)), weighted_q2(size(pdf1,1))
    !----------------------------
    integer i
    if (cs%proc /= "DY") call wae_error("lumi_DY_qg: should not be called with proc != DY")

    weighted_q1 = zero
    weighted_q2 = zero
    do i = 1, nf_int
       weighted_q1 = weighted_q1 + gv2_ga2(i) * (pdf1(:,-i) + pdf1(:,i))
       weighted_q2 = weighted_q2 + gv2_ga2(i) * (pdf2(:,-i) + pdf2(:,i))
    end do

    res = (  PartonLuminosity(grid, weighted_q1(:), pdf2(:,0)) &
         & + PartonLuminosity(grid, weighted_q2(:), pdf1(:,0)) ) * cs%norm_fb
  end function inclusive_lumi_DY_qg


end module pdfs_tools_extra_routines
