!========================================================
!--------------------------------------------------------
! Includes a module providing the resummation ingredients
! -------------------------------------------------------
!========================================================
module rad_tools
  use types; use consts_dp
  use hoppet_v1
  use ew_parameters
  use mass_corr
  implicit none

  private

  ! the following flag decides whether to perform a channel breakdown
  logical, parameter, public :: flavour_breakdown=.false.

  ! define the orders that we will support
  integer, parameter, public :: order_LL = 0
  integer, parameter, public :: order_NLL = 1
  integer, parameter, public :: order_NNLL = 2
  integer, parameter, public :: order_NNNLL = 3

  ! a type to conveniently hold the coupling and information on
  ! scale ratios
  type process_and_parameters
     character(len=5) :: collider, proc
     real(dp) :: rts
     real(dp) :: alphas_muR, as2pi
     real(dp) :: M, muR, muF, Q, p
     real(dp) :: ln_muR2_M2
     real(dp) :: ln_Q2_M2
     real(dp) :: ln_Q2_muR2
     real(dp) :: ln_muF2_M2
     real(dp) :: M2_rts2
     real(dp) :: jet_radius, jet_veto
     real(dp) :: m_min, m_max, eta_min, eta_max
     real(dp) :: xmuR_init, xmuF_init, xQ_init !initial values of xmuR, xmuF and Q
     character(len=7) :: loop_mass
     character(len=10) :: observable
     character(len=1) :: flavour1, flavour2
     logical :: dynamical, dryrun, analytic_expansion, prime
     integer :: res_order
     real(dp) :: norm_fb
     logical :: include_c1_squared
     character(len=4) :: Mscheme ! this is the quark mass scheme (M_OS: on-shell; M_MS: MS bar)
     real(dp) ::  mu_mass_MSbar
     logical :: DY_fixedMY !this flag tells if we are running with M,Y fixed
     logical :: use_alphas_LHAPDF
     real(dp) :: fixed_M, fixed_Y !values of M and Y if only one point is needed
     logical :: numerical_jetveto
     character(len=4) :: hdecay_mode
  end type process_and_parameters


  public :: order_string
  public :: process_and_parameters, set_process_and_parameters, reset_dynamical_parameters, set_q_over_m, print_parameters, get_lambda, set_prime_counting, set_fixed_M_and_Y, use_numerical_jetveto
  public  :: init_proc, Ltilde, Ltilde_scalar
  real(dp), public :: A(4), B(3), as_pow
  real(dp), public :: CC, BB ! colour factor & our B
  real(dp), public :: H(2), Hm(1)
  real(dp), public :: H1, Hm1, H2
  type(process_and_parameters), public, save :: cs

contains

  !======================================================================
  function order_string(order) result(res)
    integer,    intent(in) :: order
    character(len=4) :: res
    !------------------------------------
    select case(order)
    case(order_LL)
       res = "LL"
    case(order_NLL)
       res = "NLL"
    case(order_NNLL)
       res = "NNLL"
    case(order_NNNLL)
       res = "NNNLL"
    case default
       call wae_error("value of order was not recognized in order_string",intval=order)
    end select
  end function order_string

  !======================================================================
  subroutine init_proc()
    implicit none
  !======================================================================
    real(dp) :: Lt, Gamma_4
    real(dp), parameter :: zeta4=pi**4/90._dp, zeta5=1.0369277551433699263313654864570

    select case(trim(cs%proc))
       !! B(2) determined from Grazzini, De Florian
       !! hep-ph/0108273, with some additional manipulations
       !! reflecting that in our convention H(1) multiplies alphas(muR).
       !!
       !! A(3) from Becher, Neubert 1007.4005 [hep-ph]
       !!
       !! The explicit formulae are in section 1 of the supplementary
       !! material of BMSZ.
    case('H')
       as_pow = two
       A(1) = two*ca_def
       CC   = ca_def
       A(2) = cmw_K*A(1)
       A(3) = cmw_K2*A(1)+pi*beta0*ca_def*(ca_def*(808._dp/27._dp-28._dp*zeta3)-224._dp/27._dp*tf_def)
       BB   = -beta0 * pi/ca_def
       B(1) = -two*twopi*beta0
       B(2) = -two*(ca_def**2*(8._dp/3._dp+three*zeta3)-cf_def*tf_def-four/three*ca_def*tf_def) &
            & +twopi_beta0*zeta2*ca_def
       ! B(3), A(4) extracted from 1604.01404, using eq. (71) of 1007.4005 (we set the four-loop cusp Gamma_4 to zero)
       B(3) = (-270*cf_def*nf*(9*cf_def + 11*nf) + &
     &    ca_def*nf*(5*nf*(529 + 60*pi**2 - 144*zeta3) - &
     &       27*cf_def*(-1205 + 30*pi**2 + 8*pi**4 - 720*zeta4)) + &
     &    ca_def**2*nf*(-53185 - 5990*pi**2 + 738*pi**4 + 12360*zeta2 + 72720*zeta3 - 8100*zeta4) + &
     &    ca_def**3*(-2871*pi**4 + pi**2*(30545 - 5400*zeta3) + &
     &       5*(34219 - 91188*zeta3 + 24*zeta2*(-799 + 594*zeta3) - 12474*zeta4 + 38880*zeta5)))/9720.
       ! four-loop cusp from 1805.09638, eq. 4.2
       Gamma_4 = 40880 - 11714*nf + 440.0488*nf**2 + 7.362774*nf**3
       Gamma_4 = Gamma_4 * 2**4/4**4 ! conversion from Moch's notation
       A(4) = Gamma_4 + (ca_def*pi*(-288*beta1*pi*(28*nf + ca_def*(-202 + 189*zeta3)) + &
            &      beta0*(2*ca_def*nf*(-31313 + 3708*zeta2 + 12204*zeta3 - 2430*zeta4) + &
            &         nf*(32*nf*(58 + 81*zeta3) + 27*cf_def*(-1711 + 912*zeta3 + 432*zeta4)) + &
            &         ca_def**2*(297029 - 332856*zeta3 + 72*zeta2*(-799 + 594*zeta3) - 37422*zeta4 + &
            &            139968*zeta5))))/972.
       H1 = ca_def*(5._dp+7._dp/6._dp*pisq)-three*cf_def  ! in the Mtop->infty limit
       if (cs%loop_mass .ne. 'none') then
          if (cs%Mscheme.ne.'M_OS'.and.cs%Mscheme.ne.'M_MS') call wae_error('heavy-quark mass corrections:  unrecognized mass scheme: '//cs%Mscheme)
          select case(trim(cs%loop_mass))
          case('t')
             if (cs%Mscheme=='M_OS') then
                ! On shell scheme
                Hm1 = two*c1h(cs%M, mt, mb, mc, ms, case='t', scheme=cs%Mscheme)   + half*pisq*CC
             else
                ! MS bar
                if (cs%mu_mass_MSbar.eq.-one) then
                   Hm1 = two*c1h(cs%M, cs%muR, cs%muR, cs%muR, cs%muR, case='t', scheme=cs%Mscheme)   + half*pisq*CC
                else
                   Hm1 = two*c1h(cs%M, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, case='t', scheme=cs%Mscheme)   + half*pisq*CC
                end if
             end if
          case('b')
             if (cs%Mscheme == 'M_OS') then
                Hm1 = two*c1h(cs%M, mt, mb, mc, ms, case='b', scheme=cs%Mscheme)   + half*pisq*CC
             else
                if (cs%mu_mass_MSbar.eq.-one) then
                   Hm1 = two*c1h(cs%M, cs%muR, cs%muR, cs%muR, cs%muR, case='b', scheme=cs%Mscheme)   + half*pisq*CC
                else
                   Hm1 = two*c1h(cs%M, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, case='b', scheme=cs%Mscheme)   + half*pisq*CC
                end if
             end if
          case('t+b')
             if (cs%Mscheme == 'M_OS') then
                Hm1 = two*c1h(cs%M, mt, mb, mc, ms, case='t+b', scheme=cs%Mscheme) + half*pisq*CC
             else
                if (cs%mu_mass_MSbar.eq.-one) then
                   Hm1 = two*c1h(cs%M,  cs%muR, cs%muR, cs%muR, cs%muR, case='t+b', scheme=cs%Mscheme) + half*pisq*CC
                else
                   Hm1 = two*c1h(cs%M,  cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, case='t+b', scheme=cs%Mscheme) + half*pisq*CC
                end if
             end if
          case('t+b+c')
             if (cs%Mscheme == 'M_OS') then
                Hm1 = two*c1h(cs%M, mt, mb, mc, ms, case='t+b+c', scheme=cs%Mscheme) + half*pisq*CC
             else
                if (cs%mu_mass_MSbar.eq.-one) then
                   Hm1 = two*c1h(cs%M,  cs%muR, cs%muR, cs%muR, cs%muR, case='t+b+c', scheme=cs%Mscheme) + half*pisq*CC
                else
                   Hm1 = two*c1h(cs%M,  cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, case='t+b+c', scheme=cs%Mscheme) + half*pisq*CC
                end if
             end if
          case('t+b+c+s')
             if (cs%Mscheme == 'M_OS') then
                Hm1 = two*c1h(cs%M, mt, mb, mc, ms, case='t+b+c+s', scheme=cs%Mscheme) + half*pisq*CC
             else
                if (cs%mu_mass_MSbar.eq.-one) then
                   Hm1 = two*c1h(cs%M,  cs%muR, cs%muR, cs%muR, cs%muR, case='t+b+c+s', scheme=cs%Mscheme) + half*pisq*CC
                else
                   Hm1 = two*c1h(cs%M,  cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, case='t+b+c+s', scheme=cs%Mscheme) + half*pisq*CC
                end if
             end if
          case default
             call wae_error('heavy-quark mass corrections:  unrecognized option: '//cs%loop_mass)
          end select
          Hm(1) = Hm1 + (-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2   &
               & +two*as_pow*pi*beta0*cs%ln_muR2_M2
       endif
       H(1) = H1 + (-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2   &
            & +two*as_pow*pi*beta0*cs%ln_muR2_M2

       ! H(2) amounts to the two-loop virtual corrections,
       ! the constant term of the Sudakov (see jatRad.nb) and renormalisation
       ! of lower order virtual corrections
       ! two-loop virtual corrections from Becher-Neubert 1205.3806 (obtained with nf=5, N=3)
       Lt =  log(mH**2/mt**2)
       H2   =  5359._dp/54._dp + (137._dp*Lt)/6._dp + (1679._dp*pi**2)/24._dp &
            & + (37._dp*pi**4)/8._dp - (499._dp*zeta3)/6._dp
       if (cs%loop_mass .ne. 'none') then
          H(2) =  H2 + (A(1)*B(1)*half+A(1)*pi*beta0/three)*(-cs%ln_Q2_M2)**3   &
               & + 1/8._dp*A(1)**2*(-cs%ln_Q2_M2)**4 + (half*(-A(2)+B(1)**2-A(1)*Hm1) &
               & + pi*beta0*(B(1)-A(1)*(-cs%ln_Q2_muR2)))*cs%ln_Q2_M2**2 &
               & + (-B(2)-B(1)*(Hm1 + two*pi*beta0*(-cs%ln_Q2_muR2)))*(-cs%ln_Q2_M2)        &
               ! add the renormalisation scale dependence
               & - as_pow*four*(-half*(one+as_pow)*pisq*beta0**2*cs%ln_muR2_M2**2-pisq*beta1*cs%ln_muR2_M2) &
               & + Hm1*two*(one+as_pow)*pi*beta0*cs%ln_muR2_M2 &
               ! the next term is the cross product of the two scale dependences on H1
               & + two*as_pow*pi*beta0*cs%ln_muR2_M2*(-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2
       else
          H(2) =  H2 + (A(1)*B(1)*half+A(1)*pi*beta0/three)*(-cs%ln_Q2_M2)**3   &
               & + 1/8._dp*A(1)**2*(-cs%ln_Q2_M2)**4 + (half*(-A(2)+B(1)**2-A(1)*H1) &
               & + pi*beta0*(B(1)-A(1)*(-cs%ln_Q2_muR2)))*cs%ln_Q2_M2**2 &
               & + (-B(2)-B(1)*(H1 + two*pi*beta0*(-cs%ln_Q2_muR2)))*(-cs%ln_Q2_M2)        &
               ! add the renormalisation scale dependence
               & - as_pow*four*(-half*(one+as_pow)*pisq*beta0**2*cs%ln_muR2_M2**2-pisq*beta1*cs%ln_muR2_M2) &
               & + H1*two*(one+as_pow)*pi*beta0*cs%ln_muR2_M2  &
               ! the next term is the cross product of the two scale dependences on H1
               & + two*as_pow*pi*beta0*cs%ln_muR2_M2*(-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2
          ! add the additional term from our Sudakov
          if (.not.cs%analytic_expansion) H(2) = H(2) + 8._dp*A(1)*pi*beta0/3._dp*zeta3
        endif
    case('DY','Wp','Wm')
       as_pow = zero
       A(1) = two*cf_def
       CC   = cf_def
       A(2) = cmw_K*A(1)
       A(3) = cmw_K2*A(1)+pi*beta0*cf_def*(ca_def*(808._dp/27._dp-28._dp*zeta3)-224._dp/27._dp*tf_def)
       BB   = -three/four
       B(1) = -three*cf_def
       B(2) =   -two*(cf_def**2*(-half*pisq+3._dp/8._dp+6._dp*zeta3)&
            & + cf_def*ca_def*(11._dp/18._dp*pisq+17._dp/24._dp-three*zeta3) &
            & + cf_def*tf_def*(-one/6._dp-two/9._dp*pisq)) &
            & + twopi_beta0*zeta2*cf_def
       ! B(3), A(4) extracted from 1604.01404, using eq. (71) of 1007.4005 (we set the four-loop cusp Gamma_4 to zero)
       B(3) = ((ca_def*cf_def*nf*(-47.51165980795611 + (5188*Pi**2)/243. + (44*Pi**4)/45. - (3856*zeta3)/27.))/2. - &
            &    cf_def**2*nf*(63.370370370370374 - (8*Pi**4)/45. - (304*zeta3)/9.) - &
            &    ca_def*cf_def*nf*(85.90672153635117 - (412*Pi**2)/243. + (2*Pi**4)/27. - (904*zeta3)/27.) - cf_def*nf**2*(-2.5459533607681757 - (32*zeta3)/9.) + &
            &    (cf_def*nf**2*(26.52400548696845 - (80*Pi**2)/27. - (64*zeta3)/27.))/4. + &
            &    (cf_def**2*nf*(218.74074074074073 - (52*Pi**2)/9. - (56*Pi**4)/27. + (1024*zeta3)/9.))/2. + &
            &    ca_def**2*cf_def*(-95.5727023319616 - (7163*Pi**2)/243. - (83*Pi**4)/45. + (7052*zeta3)/9. - (88*Pi**2*zeta3)/9. - 272*zeta5) + &
            &    ca_def*cf_def**2*(-75.5 + (410*Pi**2)/9. + (494*Pi**4)/135. - (1688*zeta3)/3. - (16*Pi**2*zeta3)/3. - 240*zeta5) - &
            &    ca_def**2*cf_def*(-407.4471879286694 + (3196*Pi**2)/243. + (77*Pi**4)/135. + (12328*zeta3)/27. - (88*Pi**2*zeta3)/9. - 192*zeta5) + &
            &    cf_def**3*(-29 - 6*Pi**2 - (16*Pi**4)/5. - 136*zeta3 + (32*Pi**2*zeta3)/3. + 480*zeta5))/8.
       ! four-loop cusp from 1805.09638, eq. 4.1
       Gamma_4 = 20702 - 5171.9*nf + 195.5772*nf**2 + 3.272344*nf**3
       Gamma_4 = Gamma_4 * 2**4/4**4 ! conversion from Moch's notation
       A(4) = Gamma_4 - (cf_def*pi*(288*beta1*pi*(28*nf + ca_def*(-202 + 189*zeta3)) - &
            &       beta0*(2*ca_def*nf*(-31313 + 3708*zeta2 + 12204*zeta3 - 2430*zeta4) + &
            &          nf*(32*nf*(58 + 81*zeta3) + 27*cf_def*(-1711 + 912*zeta3 + 432*zeta4)) + &
            &          ca_def**2*(297029 - 332856*zeta3 + 72*zeta2*(-799 + 594*zeta3) - 37422*zeta4 + &
            &             139968*zeta5))))/972.
       H1   = cf_def*(-8._dp+7._dp/6._dp*pisq)
       H(1) = H1 +(-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2   &
            & + two*as_pow*pi*beta0*cs%ln_muR2_M2
       H2 = (-57433d0/972d0+281d0/162d0*pi**2        &
            &           +22d0/27d0*pi**4+1178d0/27d0*zeta3)
       H(2) =  H2 + (A(1)*B(1)*half+A(1)*pi*beta0/three)*(-cs%ln_Q2_M2)**3   &
            & + 1/8._dp*A(1)**2*(-cs%ln_Q2_M2)**4 + (half*(-A(2)+B(1)**2-A(1)*H1) &
            & + pi*beta0*(B(1)-A(1)*(-cs%ln_Q2_muR2)))*cs%ln_Q2_M2**2 &
            & + (-B(2)-B(1)*(H1 + two*pi*beta0*(-cs%ln_Q2_muR2)))*(-cs%ln_Q2_M2)        &
            ! add the renormalisation scale dependence
            & - as_pow*four*(-half*(one+as_pow)*pisq*beta0**2*cs%ln_muR2_M2**2-pisq*beta1*cs%ln_muR2_M2) &
            & + H1*two*(one+as_pow)*pi*beta0*cs%ln_muR2_M2  &
            ! the next term is the cross product of the two scale dependences on H1
            & + two*as_pow*pi*beta0*cs%ln_muR2_M2*(-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2
       ! add the additional term from our Sudakov
       if (.not.cs%analytic_expansion) H(2) = H(2) + 8._dp*A(1)*pi*beta0/3._dp*zeta3
   end select
  end subroutine init_proc

  !======================================================================
  !
  subroutine set_process_and_parameters(collider, proc, rts, M, muR, muF, Q, p, res_order, mmin, mmax, etamin, etamax, &
       &jet_radius,jet_veto,loop_mass,observable,dynamical,dryrun,analytic_expansion,flavour1,flavour2, hdecay_mode, use_hoppet_running)
    character(len=*),          intent(in)  :: collider, proc, loop_mass, observable, flavour1, flavour2
    real(dp),                  intent(in)  :: rts, M, muR, muF, Q, p, jet_radius, jet_veto
    real(dp),                  intent(in)  :: mmin, mmax, etamin, etamax
    logical,                   intent(in)  :: dynamical, dryrun, analytic_expansion
    integer,                   intent(in)  :: res_order
    real(dp)                               :: alphasPDF
    character(len=*), intent(in), optional :: hdecay_mode
    logical, intent(in), optional          :: use_hoppet_running
    !-------------------------------------------------------
    cs%collider   = collider
    cs%proc       = proc
    cs%rts        = rts
    cs%M          = M
    cs%Q          = Q
    cs%muR        = muR
    cs%muF        = muF
    cs%xmuR_init  = muR/M
    cs%xmuF_init  = muF/M
    cs%xQ_init  = Q/M
    ! set this to true to use alphas from LHAPDF
    if (present(use_hoppet_running)) then
      cs%use_alphas_LHAPDF = .not. use_hoppet_running
    else
      cs%use_alphas_LHAPDF = .true.
    endif
    if (cs%use_alphas_LHAPDF) then
      cs%alphas_muR = alphasPDF(muR)
    else
      cs%alphas_muR = RunningCoupling(muR)
    endif
    cs%as2pi      = cs%alphas_muR/twopi
    cs%p          = p
    cs%ln_muR2_M2 = 2*log(muR/M)
    cs%ln_Q2_M2   = 2*log(Q/M)
    cs%ln_Q2_muR2 = 2*log(Q/muR)
    cs%ln_muF2_M2 = 2*log(muF/M)
    cs%M2_rts2    = M**2/rts**2
    cs%jet_radius = jet_radius
    cs%jet_veto = jet_veto
    cs%loop_mass  = loop_mass
    cs%observable = observable
    cs%dynamical  = dynamical
    cs%dryrun     = dryrun
    cs%analytic_expansion = analytic_expansion
    cs%numerical_jetveto = .false.
    cs%m_min      = mmin
    cs%m_max      = mmax
    cs%eta_min    = etamin
    cs%eta_max    = etamax
    cs%res_order  = res_order
    cs%flavour1    = flavour1
    cs%flavour2    = flavour2
    cs%norm_fb     = 1d0
    cs%prime       = .false.
    if(cs%proc=='H') then
      cs%norm_fb = cs%as2pi**2 * pi/144/higgs_vev**2 * invGev2_to_nb * 1e6_dp
      ! include quark mass effects if required (TODO: to be fixed)
      ! if (cs%loop_mass .ne. 'none') then
      !    ! norm_mass is the ratio of the "finite-mass" Born cross section
      !    ! to the "heavy-quark" result (checked with hnnlo-v2.0)
      !    if (mass_scheme == 'M_OS') then
      !       ! On shell scheme
      !       cs%norm_fb = cs%norm_fb * norm_mass(cs%M, mt, mb, mc, ms, cs%loop_mass, cs%Mscheme)
      !    else
      !       ! MS bar scheme
      !       if (cs%mu_mass_MSbar.eq.-one) then
      !          cs%norm_fb = cs%norm_fb * norm_mass(cs%M,  cs%muR, cs%muR, cs%muR, cs%muR, cs%loop_mass, cs%Mscheme)
      !       else
      !          cs%norm_fb = cs%norm_fb * norm_mass(cs%M,  cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%mu_mass_MSbar, cs%loop_mass, cs%Mscheme)
      !       end if
      !    end if
      elseif (cs%proc=='DY') then
        ! Eq.9.32 from ESW
        cs%norm_fb = pi/3.0_dp * sqrt(two) * GF_GeVm2 * invGev2_to_nb * 1e6_dp
      endif
      cs%include_c1_squared = .false.
      ! relevant parameter if we are interested in a single phase-space point.
      ! it is not particularly elegant but it does the job.
      ! the parameters are initialised by the routine set_fixed_M_and_Y below
      cs%DY_fixedMY = .false.
      cs%fixed_M    = 91.187
      cs%fixed_Y    = 0
      if (present(hdecay_mode)) then
         cs%hdecay_mode = hdecay_mode
      else
         cs%hdecay_mode = "none"
      endif
  end subroutine set_process_and_parameters


  !======================================================================
  !
  subroutine reset_dynamical_parameters(muR, muF, Msinglet) BIND(C)
    use iso_c_binding
    implicit none
    real(c_double), intent(in)  :: muR, muF, Msinglet
    real(dp)                    :: alphasPDF
    !-------------------------------------------------------
    cs%muR        = muR
    cs%muF        = muF
    cs%M          = Msinglet
    cs%Q          = cs%xQ_init*cs%M ! keep Q/M constant

    if (cs%use_alphas_LHAPDF) then
      cs%alphas_muR = alphasPDF(cs%muR)
    else
      cs%alphas_muR = RunningCoupling(cs%muR)
    endif
    cs%as2pi      = cs%alphas_muR/twopi
    cs%ln_Q2_muR2 = 2*log( cs%Q   / cs%muR )
    cs%ln_muF2_M2 = 2*log( cs%muF / cs%M)
    cs%ln_muR2_M2 = 2*log( cs%muR / cs%M)
    cs%ln_Q2_M2   = 2*log( cs%Q   / cs%M)
    cs%M2_rts2    = cs%M**2/cs%rts**2

    call init_proc() !-- recalculates with new muR,muF

  end subroutine reset_dynamical_parameters


  !======================================================================
  !
  subroutine set_q_over_m(xQ) BIND(C)
    use iso_c_binding
    implicit none
    real(c_double), intent(in)  :: xQ
    !-------------------------------------------------------
    cs%xQ_init = xQ
  end subroutine set_q_over_m


  !======================================================================
  !
  subroutine set_prime_counting() BIND(C)
   implicit none
   !-------------------------------------------------------
   cs%prime = .true.
  end subroutine set_prime_counting

  !======================================================================
  !
  subroutine use_numerical_jetveto() BIND(C)
   implicit none
   !-------------------------------------------------------
   cs%numerical_jetveto = .true.
 end subroutine use_numerical_jetveto


  !======================================================================
  !
  subroutine set_fixed_M_and_Y(M,Y) BIND(C)
   implicit none
   real(dp), intent(in)  :: M, Y
   !-------------------------------------------------------
   cs%DY_fixedMY = .true.
   cs%fixed_M    = M
   cs%fixed_Y    = Y
end subroutine set_fixed_M_and_Y


  !======================================================================
  subroutine print_parameters(iunit)
    integer, intent(in) :: iunit
    real(dp)                    :: alphasPDF

    write(iunit,'(a)') "# process = "//cs%proc
    write(iunit,'(a)') "# collider = "//cs%collider
    write(iunit,'(a,f10.4)') "# rts = ", cs%rts
    write(iunit,'(a,f10.4)') "# M   = ", cs%M
    if(cs%dynamical)then
       write(iunit,'(a)') "# muR = DYNAMICAL"
       write(iunit,'(a)') "# muF = DYNAMICAL"
    else
       write(iunit,'(a,f10.4)') "# muR = ", cs%muR
       write(iunit,'(a,f10.4)') "# muF = ", cs%muF
    endif
    write(iunit,'(a,f10.4)') "# Q   = ", cs%Q
    write(iunit,'(a,f10.4)') "# p   = ", cs%p
    write(iunit,'(a)') "# loop_mass  = "//cs%loop_mass
    if (cs%use_alphas_LHAPDF) then
      write(iunit,'(a,f10.5,f10.5,a)') '# alphas(MZ) = ', alphasPDF(MZ)
      write(iunit,'(a,f10.5,f10.5,a)') '# alphas(M) = ', alphasPDF(cs%M)
      write(iunit,'(a,f10.5,f10.5,a)') '# alphas(muR) = ', alphasPDF(cs%muR)
    else
      write(iunit,'(a,f10.5,f10.5,a)') '# alphas(MZ) = ', RunningCoupling(MZ)
      write(iunit,'(a,f10.5,f10.5,a)') '# alphas(M) = ', RunningCoupling(cs%M)
      write(iunit,'(a,f10.5,f10.5,a)') '# alphas(muR) = ', RunningCoupling(cs%muR)
    endif
    write(iunit,'(a,f10.3)') "# Rjet = ", cs%jet_radius
  end subroutine print_parameters


  !=========================================================
  ! Computes the modified logarithm
  function Ltilde(v,p) result(res)
    real(dp), intent(in) :: v(:), p
    real(dp) :: Z(size(v))
    real(dp) :: res(size(v))

    if (p > 0) then
       ! Ltilde is defined setting ptmax->infty
       Z = one
       res = 1/p*log(1/v**p+1)*Z
    else
       res = log(1/v)
       !write(0,*) 'Warning: using unmodified logarithm'
       !res = zero
       !where (v<=one) res = log(1/v)
    end if

  end function Ltilde


  function Ltilde_scalar(v,p) result(res)
    real(dp), intent(in) :: v, p
    real(dp) :: Z
    real(dp) :: res

    if (p > 0) then
       ! Ltilde is defined setting ptmax->infty
       Z = one
       res = 1/p*log(1/v**p+1)*Z
    else
       res = log(1/v)
       !write(0,*) 'Warning: using unmodified logarithm'
       !res = zero
       !if (v<=one) res = log(1/v)
    end if

  end function Ltilde_scalar

  !======================================================================
  function get_lambda(L) result(res)
    real(dp),                  intent(in) :: L(:)
    !-----------------------------------------
    real(dp):: res(size(L))

    res = cs%alphas_muR * L * beta0
  end function get_lambda

end module rad_tools
